/*
 * elspy
 *
 * Exim local_scan() to embed a Python interpreter and call
 * a Python local_scan() function.
 * 
 * Copyright (c) 2002 Gregory P. Ward.  All rights reserved.
 * See README.txt for licensing information.
 *
 * The elspy home page is at http://elspy.sourceforge.net/ .
 */

#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <unistd.h>

#include <Python.h>
#include "local_scan.h"


/* If true, elspy debug messages are written to Exim's main log.
 * Otherwise, they are written to stdout.
 */
#define LOG_DEBUG_MESSAGES 1

/* If true, elspy warning and error messages (eg. can't load module,
 * Python tracebacks, etc.) are written to Exim's main log.  Otherwise
 * they are written to stderr.
 */
#define LOG_ERROR_MESSAGES 1

/* Set this higher for more elspy debug messages; set it to zero
 * for none.  Maximum debug level is currently 5.
 */
#define DEBUG_LEVEL 0


static char * __revision__ = "$Id: elspy.c,v 1.20 2002/10/21 22:19:02 gward Exp $";


/* ----------------------------------------------------------------------
 * Internal logging interface.
 *
 * These functions are just used internally by elspy -- they're
 * not exposed to Python code.
 */

static void
_log_message (char * head,
              char * format,
              va_list arglist,
              int use_exim_log,
              int exim_log,
              FILE * file)
{
   int         len;                     /* length of 'head' */
   static char msg[1024];

   strcpy(msg, head);
   len = strlen(msg);
   vsnprintf(msg+len, 1024-len, format, arglist);

   if (use_exim_log) {
      log_write(0, exim_log, msg);
   }
   else {
      fputs(msg, file);
      fputc('\n', file);
   }
}


static void
log_debug (int threshold, char * format, ...)
{
   va_list arglist;
   char prefix[30];

   va_start(arglist, format);
   if (DEBUG_LEVEL >= threshold) {
      snprintf(prefix, 30, "elspy pid %d: ", getpid());
      _log_message(prefix, format, arglist,
                   LOG_DEBUG_MESSAGES, LOG_MAIN, stdout);
   }
   va_end(arglist);
}

static void
log_error (char * format, ...)
{
   va_list arglist;

   va_start(arglist, format);
   _log_message("elspy: ", format, arglist,
                LOG_ERROR_MESSAGES, LOG_MAIN, stderr);
   va_end(arglist);
}


/* ----------------------------------------------------------------------
 * External interface
 *
 * These functions are exported to Python code -- specifically
 * they go in the '_elspy' pseudo-module.
 */

static PyObject *
expand_string_py (PyObject * self,
                  PyObject * args)
{
   char * string;

   if (!PyArg_ParseTuple(args, "s", &string)) {
      return NULL;
   }
   log_debug(3, "expand_string_py: input string = >%s<", string);
   string = (char *) expand_string((uschar *) string);
   log_debug(3, "expand_string_py: output string = >%s<", string);
   if (string == NULL)
      return Py_None;
   else
      return PyString_FromString(string);
}

static PyObject *
header_add_py (PyObject * self,
               PyObject * args)
{
   char * text;

   if (!PyArg_ParseTuple(args, "s", &text)) {
      return NULL;
   }
   header_add((int) ' ', text);
   return Py_None;
}


static PyObject *
log_write_py (PyObject * self,
              PyObject * args)
{
   int which;                           /* which Exim log to write to */
   char * msg;                          /* message to write */

   if (!PyArg_ParseTuple(args, "is", &which, &msg)) {
      return NULL;
   }
   log_write(0, which, msg);   
   return Py_None;
}


/* ----------------------------------------------------------------------
 * Bureaucracy for the external interface.
 *
 * Code/data to create and populate the '_elspy' module.
 */

static PyMethodDef elspy_methods[] = {
   {"expand_string", expand_string_py, METH_VARARGS,
    "Pass a string through Exim's string expansion code."},
   {"header_add", header_add_py, METH_VARARGS,
    "Add a header line to the current message."},
   {"log_write", log_write_py, METH_VARARGS,
    "Write a line to one of Exim's log files."},
   {NULL, NULL, 0, NULL}
};


static void
init_elspy (void)
{
   PyObject * mod;
   PyObject * mod_dict;

   mod = Py_InitModule("_elspy", elspy_methods);

   /* Add necessary constants to module's dictionary. */
   mod_dict = PyModule_GetDict(mod);
   PyDict_SetItemString(mod_dict, "LOCAL_SCAN_ACCEPT",
                        PyInt_FromLong((long) LOCAL_SCAN_ACCEPT));
   PyDict_SetItemString(mod_dict, "LOCAL_SCAN_REJECT",
                        PyInt_FromLong((long) LOCAL_SCAN_REJECT));
   PyDict_SetItemString(mod_dict, "LOCAL_SCAN_TEMPREJECT",
                        PyInt_FromLong((long) LOCAL_SCAN_TEMPREJECT));

   PyDict_SetItemString(mod_dict, "LOG_MAIN",
                        PyInt_FromLong((long) LOG_MAIN));
   PyDict_SetItemString(mod_dict, "LOG_REJECT",
                        PyInt_FromLong((long) LOG_REJECT));
}


/* ----------------------------------------------------------------------
 * Internals
 *
 * The guts of elspy -- convert Exim's data structures into
 * objects usable by Python.
 */

#define check(val) if ((val) == NULL) return NULL


static PyObject *
build_header_list (void)
{
   PyObject * headers;
   PyObject * p_header;
   header_line * cur;

   /* Exim provides a linked list of header_line structs, each of which
      contains:
        int      type     (a printable char)
        int      slen     (length of header line)
        uschar * text     (text of the header)

      We convert that to a list of 2-tuples:
        (type : int, text : string)
   */

   headers = PyList_New(0);
   cur = header_list;                   /* Exim global */
   while (cur) {
      p_header = Py_BuildValue("is", cur->type, (char *) cur->text);
      PyList_Append(headers, p_header);
      /*Py_DECREF(p_header);*/
      cur = cur->next;
   }

   return headers;

} /* build_header_list () */


static PyObject *
build_info_dict (void)
{
   PyObject * info;                     /* dict to return */
   PyObject * val;                      /* each value of the dict in turn */
   PyObject * recip_list;
   int        i;

   info = Py_BuildValue(
       "{s:s,s:s,s:i,s:s,s:s,s:s,s:s,s:i}",
       "sender_address", (char *) sender_address,
       "interface_address", (char *) interface_address,
       "interface_port", interface_port,
       "received_protocol", (char *) received_protocol,
       "sender_host_address", (char *) sender_host_address,
       "sender_host_authenticated", (char *) sender_host_authenticated,
       "sender_host_name", (char *) sender_host_name,
       "sender_host_port", sender_host_port);
   if (info == NULL) {
      return NULL;
   }

   /* The recipients list is a tad more interesting. */
   log_debug(5, "info dict: building recipients_list (%d recips)", recipients_count);
   check(recip_list = PyList_New(recipients_count));
   for (i = 0; i < recipients_count; i++) {
      val = PyString_FromString((char *) recipients_list[i].address);
      check(val);
      PyList_SET_ITEM(recip_list, i, val);
   }
   log_debug(5, "info dict: adding recipients_list");
   PyDict_SetItemString(info, "recipients_list", recip_list);
   Py_DECREF(recip_list);

   return info;

} /* build_info_dict () */


static PyObject *
lookup_func (PyObject * module,
             char *     func_name,
             char *     mod_name)
{
   PyObject * mod_dict;
   PyObject * func;

   mod_dict = PyModule_GetDict(module);
   func = PyDict_GetItemString(mod_dict, func_name);

   if (func) {
      if (! PyCallable_Check(func)) {
         log_error("error: %s object in %s module not callable",
                   func_name, mod_name);
         return NULL;
      }
      else
         return func;
   }
}


static PyObject *
find_local_scan (PyObject * elspy_mod,
                 PyObject * local_mod)
{
   PyObject * lscan_func;

   /* The rules:
      - if _local_scan() exists in exim_local_scan module,
        use it -- ignore the one in elspy_mod
      - otherwise, dig up _local_scan() from elspy module
      - if that fails, we're in trouble -- the locally-installed
        elspy module is broken.  Log an error message and return
        NULL; caller will return "temporary reject" to Exim.
      - if we find something called '_local_scan' in either
        module that isn't callable, log an error message
        and return NULL.
   */

   log_debug(3, "looking up _local_scan() function in exim_local_scan module");
   lscan_func = lookup_func(local_mod, "_local_scan", "exim_local_scan");
   if (lscan_func)
      return lscan_func;
                                
   log_debug(3, "_local_scan() not defined in exim_local_scan module -- "
             "trying elspy module");
   lscan_func = lookup_func(elspy_mod, "_local_scan", "elspy");
   if (lscan_func)
      return lscan_func;
   else {
      log_error("error: _local_scan() function not found in elspy module");
      return NULL;
   }      
}


static char *
format_exception ()
{
   PyObject *type, *value, *traceback;
   PyObject *exc_name;                  /* '__name__' attr of type */
   PyObject *exc_msg;                   /* str(value) */
   int      msg_len;
   uschar * msg;                        /* complete message */

   PyErr_Fetch(&type, &value, &traceback);
   check(exc_name = PyObject_GetAttrString(type, "__name__"));
   check(exc_msg = PyObject_Str(value));
   
   msg_len = PyString_Size(exc_msg);
   if (msg_len == 0)
      msg = string_copy((uschar *) PyString_AS_STRING(exc_name));
   else
      msg = string_sprintf("%s: %s",
                           (uschar *) PyString_AS_STRING(exc_name),
                           (uschar *) PyString_AS_STRING(exc_msg));

   /* so we can PyErr_Print() a full traceback to stderr. */
   PyErr_Restore(type, value, traceback);

   return msg;
}


#if (DEBUG_LEVEL > 0)
static void
dump_recips_list ()
{
   int i;

   for (i = 0; i < recipients_count; i++) {
      printf("  %x:%s\n",
             recipients_list[i].address,
             recipients_list[i].address);
      /*if (i < recipients_count-1)
         putchar(',');
      else
      putchar('\n');*/
   }
}
#endif


static int rebuild_recipients_list (PyObject * info)
{
   PyObject * recip_list;
   int        num_recip;                /* length of recip_list */
   int        i;

   recip_list = PyDict_GetItemString(info, "recipients_list");
   if (recip_list == NULL) {
      log_error("error: recipients_list removed from info dictionary");
      return 0;
   }
   if (! PyList_Check(recip_list)) {
      log_error("error: recipients_list not a list");
      return 0;
   }

   num_recip = PyList_GET_SIZE(recip_list);

   /* Loop until the old Exim list is exhausted. */
   for (i = 0; i < recipients_count; i++) {
      uschar   * p_addr;                /* current element of recip_list */
      recipient_item e_recip;
      uschar   * e_addr;
      

      if (i >= num_recip)               /* Python list exhausted */
         break;

      p_addr = (uschar *) PyString_AsString(PyList_GET_ITEM(recip_list, i));
      if (p_addr == NULL) {
         log_error("error: element %d of recipients_list not a string", i);
         return 0;
      }

      e_addr = recipients_list[i].address;
      if (strcmp(p_addr, e_addr) != 0) {
         recipients_list[i].address = string_copy(p_addr);
      }
   }

   /* Reached the end of the Python list -- truncate Exim's list. */
   if (num_recip < recipients_count) {
      recipients_count = num_recip;
   }

   /* Still elements left in the Python list -- add them to Exim's list. */
   else if (num_recip > recipients_count) {
      for (; i < num_recip; i++) {
         uschar * p_addr;

         p_addr = (uschar *) PyString_AsString(PyList_GET_ITEM(recip_list, i));
         if (p_addr == NULL) {
            log_error("error: element %d of recipients_list not a string", i);
            return 0;
         }
         receive_add_recipient(p_addr, -1);
      }
   }

   return 1;
}


static int py_initialized = 0;

int
local_scan (int fd, uschar **return_text)
{
   PyObject * elspy_mod;                /* the elspy module */
   PyObject * local_mod;                /* the exim_local_scan module */
   PyObject * lscan_func;
   PyObject * headers;
   PyObject * info;
   PyObject * p_fd;
   PyObject * lscan_args;
   PyObject * lscan_return;
   int        result;

   if (! py_initialized) {
      log_debug(2, "initializing Python interpreter and _elspy module");
      Py_Initialize();
      init_elspy();
      py_initialized = 1;
   }
   else {
      log_debug(3, "Python interpreter already initialized");
   }

   /* Import the elspy package. */
   log_debug(1, "attempting to import module 'elspy'");
   elspy_mod = PyImport_ImportModule("elspy");
   if (elspy_mod == NULL) {
      uschar * error;

      log_error("error importing module 'elspy' "
                "(accepting message; exception follows)");
      error = format_exception();
      if (error)
         log_error(error);
      else
         log_error("unable to format Python exception");

      /* Uncomment this to get a traceback.  Should only do this in
         development mode -- if Exim is running as a daemon, the
         traceback winds up going to the SMTP client! */
      /*PyErr_Print();*/
      return LOCAL_SCAN_ACCEPT;
   }

   /* Attempt to import the exim_local_scan module.  If that fails,
      then this installation is incomplete -- just log a warning
      and accept the message. */
   log_debug(1, "attempting to import module 'exim_local_scan'");
   local_mod = PyImport_ImportModule("exim_local_scan");
   if (local_mod == NULL) {
      log_error("error importing module 'exim_local_scan' "
                "(accepting message; traceback follows)");
      PyErr_Print();
      return LOCAL_SCAN_ACCEPT;
   }

   /* Look for _local_scan() first in the locally-supplied
      exim_local_scan module, then in the standard elspy module. */
   lscan_func = find_local_scan(elspy_mod, local_mod);
   if (lscan_func == NULL)
      return LOCAL_SCAN_TEMPREJECT;

   /* OK, we have a function to call.  Start building its arg tuple,
      starting with the list of headers. */
   log_debug(3, "building header list");
   headers = build_header_list();
   if (headers == NULL) {
      log_error("error: could not build header list (traceback follows)");
      PyErr_Print();
      return LOCAL_SCAN_TEMPREJECT;
   }

   /* Build dictionary of miscellaneous info. */
   log_debug(3, "building info dict");
   info = build_info_dict();
   if (info == NULL) {
      log_error("error: could not build info dictionary (traceback follows)");
      PyErr_Print();
      return LOCAL_SCAN_TEMPREJECT;
   }

   /* Construct argument tuple: (info, headers, fd) */
   log_debug(3, "constructing arg tuple");
   lscan_args = PyTuple_New(3);         /* XXX need decref? */
   p_fd = PyInt_FromLong(fd);           /* no decref - stolen by SetItem */
   if (lscan_args == NULL || p_fd == NULL) {
      log_error("error creating arg tuple");
      return LOCAL_SCAN_TEMPREJECT;
   }
   Py_INCREF(info);
   PyTuple_SET_ITEM(lscan_args, 0, info);
   PyTuple_SET_ITEM(lscan_args, 1, headers);
   PyTuple_SET_ITEM(lscan_args, 2, p_fd);

   /* Call the Python _local_scan() function */
   log_debug(2, "calling _local_scan() function");
   lscan_return = PyObject_CallObject(lscan_func, lscan_args);
   if (lscan_return == NULL) {
      log_error("error: call to _local_scan() function failed "
                "(traceback follows)");
      PyErr_Print();
      return LOCAL_SCAN_TEMPREJECT;
   }
   Py_DECREF(lscan_args);

   /* Typecheck and interpret the return value: must be
    * either a lone int or an (int, string) tuple.
    */
   log_debug(2, "typechecking and interpreting return value");

   if (PyInt_Check(lscan_return)) {
      result = (int) PyInt_AS_LONG(lscan_return);
   }
   else if (PyTuple_Check(lscan_return) &&
            PyTuple_GET_SIZE(lscan_return) == 2) {

      PyObject * p_result;
      PyObject * p_message;

      /* both are borrowed refs */
      p_result = PyTuple_GET_ITEM(lscan_return, 0);
      p_message = PyTuple_GET_ITEM(lscan_return, 1);

      if (! (PyInt_Check(p_result) &&
             (PyString_Check(p_message) || p_message == Py_None))) {
         log_error("error: return value of _local_scan() must be "
                   "(int, string) tuple");
         return LOCAL_SCAN_TEMPREJECT;
      }
      result = (int) PyInt_AS_LONG(p_result);

      if (p_message != Py_None) 
         /* Exim must not dealloc this pointer -- it points right into
            Python's string object! */
         *return_text = (uschar *) PyString_AS_STRING(p_message);
   }
   else {
      log_error("error: return value of _local_scan() must be "
                "either an int or 2-tuple");
      return LOCAL_SCAN_TEMPREJECT;
   }

   /* Rebuild the recipients list based on changes to the list in
      the info dict passed to _local_scan(). */
   if (!rebuild_recipients_list(info))
      return LOCAL_SCAN_TEMPREJECT;
   Py_DECREF(info);
      
   return result;

}

#ifdef DLOPEN_LOCAL_SCAN
/* Return the verion of the local_scan ABI, if being compiled as a .so */
int
local_scan_version(void)
{
    return LOCAL_SCAN_ABI_VERSION ;
}
#endif
